<?php

namespace Roots\Sage\Form\Feature;

trait Setup {

	protected $used_features = [];

	public function setup_features() {
		$available_features = [
			'Expiry',
			'Draft',
		];

		// Each trait in this array will have the full namespace
		$used_features_ns = class_uses( get_called_class(), false );

		foreach ( $available_features as $feature ) {
			// Defined in the feature trait
			if ( method_exists( $this, "{$feature}_setup" ) ) {
				$this->{"{$feature}_setup"}();
			}

			if ( in_array( __NAMESPACE__ . '\\' . $feature, $used_features_ns, true ) ) {
				$this->used_features[] = $feature;
			}
		}

		add_filter( "af/form/attributes/key={$this->form_key}", [ $this, 'form_features_attribute' ], 10, 3 );
	}

	public function form_features_attribute( $form_attributes, $form, $args ) {
		if ( ! empty( $this->used_features ) ) {
			$form_attributes['data-features'] = implode( ' ', $this->used_features );
		}
		return $form_attributes;
	}

	public function using_feature( $feature ) {
		return in_array( $feature, $this->used_features, true );
	}
}

trait Expiry {

	public $Expiry_published_status = 'publish';
	public $Expiry_expired_status = 'draft';

	/**
	 * Get the meta key of the expiry
	 * @return string Meta key where the expiry timestamp is stored
	 */
	abstract function get_expiry_key();

	public function Expiry_setup() {
		add_action( "unpublish_expired_{$this->post_type}", [ $this, 'Expiry_unpublish' ], 10, 1 );
		add_action( 'updated_postmeta',                     [ $this, 'Expiry_schedule_on_save_meta' ], 10, 4 );
		add_action( "save_post_{$this->post_type}",         [ $this, 'Expiry_schedule_on_save_post' ], 10, 1 );
	}

	public function Expiry_unpublish( $post_id ) {
		global $wpdb;

		if ( ! $post = get_post( $post_id ) ) return;
		if ( $post->post_status != $this->Expiry_published_status ) return;

		$new_status = $this->Expiry_expired_status;
		$wpdb->update( $wpdb->posts, array( 'post_status' => $new_status ), array( 'ID' => $post->ID ) );
		clean_post_cache( $post->ID );

		$old_status = $post->post_status;
		$post->post_status = $new_status;
		wp_transition_post_status( $new_status, $old_status, $post );
	}

	public function Expiry_schedule( $post_id, $expiry ) {
		wp_clear_scheduled_hook( "unpublish_expired_{$this->post_type}", array( $post_id ) );

		$expiry = strtotime( $expiry );
		if ( $expiry === false ) return;

		wp_schedule_single_event( $expiry, "unpublish_expired_{$this->post_type}", array( $post_id ) );
	}

	public function Expiry_schedule_on_save_meta( $meta_id, $post_id, $meta_key, $meta_value ) {
		if ( $meta_key != $this->get_expiry_key() ) return;

		$post = get_post( $post_id );
		if ( empty( $post ) ) return;
		if ( $post->post_type != $this->post_type ) return;

		$this->Expiry_schedule( $post_id, $meta_value );
	}

	public function Expiry_schedule_on_save_post( $post_id ) {
		$expiry = get_field( $this->get_expiry_key(), $post_id, false );

		$this->Expiry_schedule( $post_id, $expiry );
	}

}

trait Draft {

	public function Draft_setup() {
		add_action( 'acf/validate_save_post', [ $this, 'Draft_validation' ], 9 );
	}

	public function Draft_validation() {
		if ( empty( $_POST['acf'] ) || empty( $_POST['af_form'] ) ) return;
		if ( $_POST['af_form'] != $this->form_key ) return;

		if ( $this->draft_submitted() ) {
			acf_reset_validation_errors();
			remove_action( 'acf/validate_save_post', [ $this, 'validation_setup' ] );
			return;
		}
	}

	// Helper function
	public function draft_submitted() {
		return (bool) ( $_POST['draft'] ?? false );
	}

}
