<?php

namespace Roots\Sage;

abstract class Form {
	use Form\Feature\Setup;

	private static $forms = [];

	public $url        = '/pages/';
	public $url_ep     = 'edit'; // Endpoint name ( /pages/edit/1234/ )
	public $url_ep_new = 'new'; // Endpoint value for new entries ( /pages/edit/new/ )
	public $post_type  = 'post';
	public $title      = '';
	public $title_edit = ''; // Title when editing a record
	public $content    = '';
	public $template   = null;
	public $form_key   = '';
	public $restricted = false;
	public $acf_groups = [];
	public $all_fields = [];
	public $af_form    = null;

	// Should the form also be displayed when no endpoint is present?
	public $show_on_base_url = false;

	public $virtual_page = null;

	public function __construct() {
		self::setup();

		$this->setup_features();

		add_action( 'init',                [ $this, 'register_virtual_page' ], 1 );
		add_action( 'init',                [ $this, 'register_endpoints' ], 10 );
		add_action( 'acf/init',            [ $this, 'register_acf_groups' ], 9 );
		add_action( 'acf/init',            [ $this, 'register_form' ], 10 );
		add_action( 'parse_request',       [ $this, 'update_title' ], 100, 1 );
		add_filter( 'query_vars',          [ $this, 'register_endpoint_query_vars' ], 10, 1 );
		add_action( 'template_redirect',   [ $this, 'restrict_form_view_redirect' ] );
		add_filter( 'af/form/restriction', [ $this, 'restrict_form_view_message' ], 9, 3 );
		add_filter( 'body_class',          [ $this, 'page_body_class' ], 10, 2 );
		add_action( 'wp_enqueue_scripts',  [ $this, 'enqueue_js' ] );

		add_filter( "af/form/args/key={$this->form_key}", [ $this, 'load_setup' ], 10, 2 );
		add_action( "af/form/submission/key={$this->form_key}", [ $this, 'submit' ], 10, 3 );
		add_action( "af/form/hidden_fields/key={$this->form_key}", [ $this, 'id_field' ], 10, 2 );
		add_action( 'acf/validate_save_post', [ $this, 'validation_setup' ] );

		self::$forms[] = $this;
	}

	public function register_form() {
		if ( !function_exists( 'af_register_form' ) ) return;

		$restrict_user = $this->restricted ? [ 'message' => 'Unauthorised' ] : false;

		$this->af_form = af_register_form( [
			'title'               => $this->title,
			'key'                 => $this->form_key,
			'display' => [
				'description'     => '',
				'success_message' => '',
			],
			'create_entries'      => true,
			'restrict_user'       => $restrict_user,
			'restrict_entries'    => false,
			'entries_limit'       => 0,
			'entries_restriction_message' => '',
		] );
	}

	public function register_acf_groups() {
		foreach ( $this->acf_groups as $group_name => $group_options ) {

			$group_options = array_merge( $group_options, [
				// Force these values
				'key'     => 'group_' . $group_name,
				'private' => true,
			] );

			$fields = $group_options['fields'];

			// Register fields explicitly below
			unset( $group_options['fields'] );

			acf_add_local_field_group( $group_options );

			foreach ( $fields as $field_name => $field_options ) {

				$field_options = array_merge( [
					'name'   => $field_name,
					'key'    => sprintf( 'field_%s_%s', $group_name, $field_name ),
					'parent' => 'group_' . $group_name,
				], $field_options );

				$field_options = self::generate_acf_sub_fields( $field_options );

				$field_options['value'] = $this->get_submitted_field_value( $field_options );

				$this->all_fields[ $field_options['key'] ] = $field_options;

				acf_add_local_field( $field_options );
			}
		}
	}

	public function register_virtual_page() {
		if ( !function_exists( 'create_virtual_page' ) ) return;

		$url_backup = $this->url;

		if ( ! $this->show_on_base_url) {
			$this->url = '/virtual-page-form/' . ltrim( $this->url, '/' );
		}

		$this->virtual_page = create_virtual_page( $this->url, [
			'url'       => $this->url,
			'title'     => $this->title,
			'content'   => $this->content,
			'template'  => $this->template,
		] );

		$this->url = $url_backup;
	}

	public function update_title( $wp ) {
		$title = ( empty( $wp->query_vars[ $this->url_ep ] ) || empty( $this->title_edit ) ) ? $this->title : $this->title_edit;

		$this->virtual_page->setTitle( $title );
		$this->af_form['title'] = $title;
	}

	public function register_endpoints() {
		static $endpoints = [];

		if ( ! in_array( $this->url_ep, $endpoints ) ) {
			$endpoints[] = $this->url_ep;

			$ep_regex = sprintf(
				'%s/%s/(.+)/?$',
				preg_quote( trim( $this->url, '/' ) ),
				preg_quote( $this->url_ep )
			);
			add_rewrite_rule( $ep_regex, 'index.php?virtual_page_url=' . urlencode( $this->virtual_page->getUrl() ) . '&' . urlencode( $this->url_ep ) . '=$matches[1]', 'top' );
		}
	}

	public function register_endpoint_query_vars( $vars ) {
		static $endpoints = [];

		if ( ! in_array( $this->url_ep, $endpoints ) ) {
			$vars[] = $this->url_ep;
		}

		return $vars;
	}

	public function is_current_page() {
		global $wp_query;
		if ( ! isset( $wp_query->virtual_page ) ) return false;
		return $wp_query->virtual_page === $this->virtual_page;
	}

	/**
	 * Define how to restrict form access
	 */
	public function restrict_form() {
		if ( $this->restricted && !is_user_logged_in() ) return true;

		return false;
	}

	public function restricted_redirect( $url = '/' ) {
		if ( is_user_logged_in() ) {
			$this->return_404();
		} else {
			wp_redirect( wp_login_url( $url ) );
			die();
		}
	}

	/**
	 * Redirect to login if restricted
	 */
	public function restrict_form_view_redirect() {
		if ( !$this->is_current_page() ) return;

		if ( $this->restrict_form() ) {
			$this->restricted_redirect( get_current_url() );
		}
	}

	/**
	 * Apply restrictions to form shortcode directly.
	 * Just in case shortcode is added to a different page.
	 */
	public function restrict_form_view_message( $restriction, $form, $args ) {
		if ( $restriction ) return $restriction;

		ob_start();
		get_template_part( 'templates/404' );
		$_404_content = ob_get_clean();

		if ( is_404() ) {
			$restriction = $_404_content;
		} else if ( $form['key'] === $this->form_key && $this->restrict_form() ) {
			if ( is_user_logged_in() ) {
				$restriction = $_404_content;
			} else {
				$restriction = sprintf( 'Please <a href="%s">log in</a> to access this page.', esc_url( wp_login_url( home_url( $this->url ) ) ) );
			}
		}

		return $restriction;
	}

	/**
	 * Called when saving an entry
	 */
	public function restrict_form_submission() {
		if ( $this->restrict_form() ) {
			$this->restricted_redirect( home_url( $this->url ) );
		}
	}

	public function get_submitted_field_value( $field ) {
		$value = null;

		if ( isset( $_POST['acf'][ $field['key'] ] ) ) {
			$value = $_POST['acf'][ $field['key'] ];
		}

		return $value;
	}

	public function validation_setup() {
		if ( !$this->is_current_page() ) return;
		if ( empty( $_POST['acf'] ) || empty( $_POST['af_form'] ) ) return;
		if ( $_POST['af_form'] != $this->form_key ) return;

		$fields = [];

		foreach ( $_POST['acf'] as $field_key => $value ) {

			$field = acf_get_field( $field_key );
			if ( !$field ) continue;

			$field['value'] = $value;
			$fields[$field['name']] = $field;
		}

		$this->validate( $this->af_form, $fields );
	}

	public function load_setup( $args, $form ) {
		$args['values'] = $this->load();
		return $args;
	}

	/**
	 * Override this in a subclass
	 */
	public function submit( $form, $fields, $args ) { }

	/**
	 * Override this in a subclass
	 */
	public function validate( $form, $fields ) {
		// $input is a field's name attribute ("acf[field_key]") (optional)
		// acf_add_validation_error( $input, $message );
	}

	/**
	 * Override this in a subclass
	 */
	public function load() {
		//return [ 'field_name' => 'foo' ];
		return [];
	}

	public function page_body_class( $classes, $additional ) {
		if ( $this->is_current_page() ) {
			$classes[] = 'af-page';
		}
		return $classes;
	}

	public function enqueue_js() {
		if ( $this->is_current_page() ) {
			acf_enqueue_scripts();
			wp_enqueue_script( 'sage/form/js' );
		}
	}

	public function return_404() {
		global $wp_query;
		$wp_query->set_404();
	}

	public function get_ep_val() {
		$val = get_query_var( $this->url_ep );
		return empty( $val ) ? null : $val;
	}

	public function id_field( $form, $args ) {
		$id = $this->get_ep_val();
		if ( ! empty( $id ) ) {
			printf( '<input type="hidden" name="id" value="%s">', esc_attr( $id ) );
		}
	}

	public function get_edit_url( $post_id = null ) {
		if ( empty( $post_id ) ) $post_id = $this->url_ep_new;
		return home_url( sprintf( '%s%s/%s/', $this->url, $this->url_ep, $post_id ) );
	}

	public function get_trash_url( $post_id = null ) {
		return wp_nonce_url( add_query_arg( 'action', 'trash', get_permalink( $post_id ) ), "trash_{$this->post_type}_{$post_id}" );
	}



	public static function setup() {
		static $setup = false;

		if ( ! $setup ) {
			$setup = true;

			self::static_hooks();
		}
	}

	public static function get_forms() {
		return self::$forms;
	}

	public static function get_form_by( $field, $search ) {
		$forms = array_merge( [], wp_list_filter( self::get_forms(), [ $field => $search ] ) );
		if ( empty( $forms ) ) return null;
		return $forms[0];
	}

	protected static function static_hooks() {
		// Remove default form handler
		add_action( 'acf/init',                 [ get_called_class(), 'remove_default_form_submit' ], 100 );

		// General form layout & style
		add_filter( 'af/form/field_attributes', [ get_called_class(), 'field_wrap_attributes' ], 10, 4 );
		add_action( 'af/form/before_fields',    [ get_called_class(), 'form_field_wrapper_start' ], 10, 2 );
		add_action( 'af/form/after_fields',     [ get_called_class(), 'form_field_wrapper_end' ], 10, 2 );
		add_filter( 'af/form/args',             [ get_called_class(), 'form_set_uploader' ], 10, 2 );

		add_action( 'wp_enqueue_scripts',       [ get_called_class(), 'register_js' ] );
	}

	public static function field_wrap_attributes( $attributes, $field, $form, $args ) {

		if ( isset( $field['prepend'] ) && $field['prepend'] != '' ) {
			$attributes['class'] .= ' acf-field-has-prepend';
		}

		if ( isset( $field['append'] ) && $field['append'] != '' ) {
			$attributes['class'] .= ' acf-field-has-append';
		}

		if ( $field['label'] == '' ) {
			$attributes['class'] .= ' acf-field-no-label';
		}

		if ( !empty( $field['ui'] ) ) {
			$attributes['class'] .= ' acf-field-has-ui';
		}

		if ( !empty( $field['class'] ) ) {
			$attributes['class'] .= ' ' . $field['class'];
		}

		$extra_data_attributes = [
			'mime_types',
			'min_width',
			'min_height',
			'min_size',
			'max_width',
			'max_height',
			'max_size',
		];

		foreach ( $extra_data_attributes as $attribute_name ) {
			if ( !empty( $field[$attribute_name] ) ) {
				$attributes["data-$attribute_name"] = $field[$attribute_name];
			}
		}

		return $attributes;
	}

	public static function form_field_wrapper_start( $form, $args ) {
		echo sprintf(
			'<div class="af-form-fields-wrap" data-form="%s">',
			esc_attr( $form['key'] )
		);
	}

	public static function form_field_wrapper_end( $form, $args ) {
		echo '</div>';
	}

	// All forms use basic uploader only
	public static function form_set_uploader( $args, $form ) {
		$args['uploader'] = 'basic';
		return $args;
	}

	public static function generate_acf_sub_fields( $parent_field ) {
		$sub_keys = [
			'sub_fields',
			'layouts',
		];

		foreach ( $sub_keys as $sub_key ) {

			if ( !empty( $parent_field[$sub_key] ) ) {
				foreach ( $parent_field[$sub_key] as $sub_field_name => $sub_field_options ) {

					$sub_field_options = acf_get_valid_field( array_merge( $sub_field_options, [
						'name'   => $sub_field_name,
						'key'    => sprintf( '%s_%s', $parent_field['key'], $sub_field_name ),
					] ) );

					$sub_field_options = self::generate_acf_sub_fields( $sub_field_options );

					$parent_field[$sub_key][$sub_field_name] = $sub_field_options;

				}
			}

		}

		return $parent_field;
	}

	public static function register_js() {
		wp_register_script( 'sage/form/js', Assets\asset_path( 'scripts/advanced-forms.js' ), [ 'jquery' ], null, true);
	}

	public static function remove_default_form_submit() {
		remove_all_actions( 'af/form/submission' );
	}
}

namespace Roots\Sage\Form;

/**
 * Useful function for creating lists of taxonomy terms
 * Can be used to populate dropdowns in ACF forms
 */
function taxonomy_field_options($taxonomy, $parent = 0, $value_field = 'term_id') {
	static $cache = [];
	$cache_key = "$taxonomy:$parent";
	if ( !array_key_exists( $cache_key, $cache ) ) {
		$terms = get_terms([
			'taxonomy' => $taxonomy,
			'orderby' => 'term_order',
			'child_of' => $parent,
			'hide_empty' => false,
		]);
		$terms = flatten_term_tree( build_term_tree( $terms, $parent ) );
		foreach ($terms as $index => $term) {
			$terms[$index]->name = str_repeat( '&nbsp;', 4 * $term->depth ) . $term->name;
			if ( !in_array( $value_field, array_keys( get_object_vars( $term ) ) ) ) {
				$terms[$index]->$value_field = get_field( $value_field, "{$term->taxonomy}_{$term->term_id}" );
			}
		}
		$cache[$cache_key] = wp_list_pluck( $terms, 'name', $value_field );
	}
	return $cache[$cache_key];
}

// Recursively move child terms into a new array property on their parent
function build_term_tree($terms, $parent = 0, $depth = 0) {
	$children = wp_list_filter($terms, ['parent' => $parent]);
	foreach ($children as $child_term) {
		$child_term->depth = $depth;
		$child_term->children = build_term_tree($terms, $child_term->term_id, $depth + 1);
	}
	return array_values($children);
}

// Remove all term children below max_depth, but keep some deep terms
function prune_term_tree($terms, $max_depth, $keep) {
	$pruned = [];
	foreach ($terms as $term) {
		$keep_term = in_array($term->term_id, $keep);
		$shallow = $term->depth <= $max_depth;
		$sibling = in_array($term->parent, $keep);
		// Only keep shallow terms that are also siblings of whitelisted terms, or whitelisted themselves
		if (($shallow && $sibling) || $keep_term) {
			$term->children = prune_term_tree($term->children, $max_depth, $keep);
			$pruned[] = $term;
		}
	}
	return $pruned;
}

// Moves child terms into their parent's array, under the parent
function flatten_term_tree($terms) {
	$flat = [];
	foreach ($terms as $term) {
		$children = flatten_term_tree($term->children);
		unset($term->children);
		$flat[] = $term;
		$flat = array_merge($flat, $children);
	}
	return $flat;
}
