<?php

namespace Roots\Sage\WooCommerce\BoxOffice;

/**
 * WooCommerce Box Office - Additional Merge Fields
 *
 * Add as many additional merge fields as you like.
 */
function extra_fields() {
  return [

    /*
    'static_example' => 'Static value',

    'dynamic_example' => function( $ticket_id ) {
      return "Ticket ID is $ticket_id";
    },
    */

    'ticket_id' => function( $ticket_id ) {
      return $ticket_id;
    },

    'product_content' => function( $ticket_id ) {
      $product_id = get_post_meta( $ticket_id, '_product', true );
      $product = wc_get_product( $product_id );

      $product_content = is_callable( [ $product, 'get_description' ] )
        ? call_user_func( [ $product, 'get_description' ] )
        : $product->post->post_content;

      $product_content = apply_filters( 'the_content', $product_content );

      return $product_content;
    },

    'ticket_html' => function( $ticket_id ) {
      // Prevent infinite loops
      static $running = false;
      if ( $running ) return '';
      $running = true;

      $css = [
        'h1' => 'margin-top: 0;
          margin-bottom: 0.5rem;
          font-family: inherit;
          font-weight: 900;
          line-height: 1.125;
          color: inherit;
          font-size: 2.25rem;',

        'p' => 'margin-top: 0;
          margin-bottom: 1rem;',

        'img' => 'vertical-align: middle;
          border-style: none;',

        'code' => 'font-family: monospace, monospace;
          font-size: 1em;',

        'wrap' => 'margin: 0;
          font-family: raleway, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, sans-serif, "Apple Color Emoji", "Segoe UI Emoji", "Segoe UI Symbol";
          font-size: 1rem;
          font-weight: 400;
          line-height: 1.5;
          color: #212529;
          text-align: left;
          background-color: #fff;
          padding-right: 20px;
          padding-left: 20px;',

        'ticket-email-wrap' => '',

        'ticket-email-header' => 'padding: 1em;
          border: 2px solid #d8d8d8;
          border-bottom: none;',

        'ticket-email-qrcode' => 'padding: 1em;
          border: 2px solid #d8d8d8;
          text-align: center;',

        'ticket-email-logo' => 'padding: 1em;
          border: 2px solid #d8d8d8;
          background-color: #d8d8d8;
          text-align: center;',

        'ticket-email-logo img' => 'width: 200px;',

        'product-content' => 'margin: 40px 0 0;',
      ];

      $css = array_map( 'esc_attr', $css );

      set_query_var( 'styles', $css );

      ob_start();
      get_template_part( 'templates/ticket-email-html' );
      $html = ob_get_clean();

      $html = wc_box_office_get_parsed_ticket_content( $ticket_id, $html );

      $running = false;
      return $html;
    },

    'barcode' => function( $ticket_id ) {
      return get_ticket_barcode_html( $ticket_id );
    },

    'ticket_type' => function( $ticket_id ) {
      $ticket = new \WC_Box_Office_Ticket( $ticket_id );

      $ticket_order_item_id = get_post_meta( $ticket_id, '_ticket_order_item_id', true );
      $ticket_order_item = new \WC_Order_Item_Product( $ticket_order_item_id );

      $variation_id = $ticket_order_item->get_variation_id();
      $variation = new \WC_Product_Variation( $variation_id );

      $variation_text = wc_get_formatted_variation( $variation, true, false );

      return $variation_text;
    },

  ];
}

/**
 * Populates the extra fields
 */
function extra_field_value( $ticket_id, $key ) {
  static $cache = [];
  static $fields = null;
  if ( is_null( $fields ) ) $fields = extra_fields();

  if ( empty( $cache[ $ticket_id ] ) ) {
    $cache[ $ticket_id ] = [];
  }

  $value = $fields[ $key ];

  if ( empty( $cache[ $ticket_id ][ $key ] ) ) {
    $cache[ $ticket_id ][ $key ] = is_callable( $value ) ? call_user_func( $value, $ticket_id ) : $value;
  }

  return $cache[ $ticket_id ][ $key ];
}

/**
 * Hooks into the admin form to display the extra fields.
 */
function extra_ticket_field_admin_helper_hook( $settings, $editor_id ) {
  if ( in_array( $editor_id, [ 'ticket-content-editor', 'ticket-email-editor' ] ) ) {
    extra_ticket_field_admin_helpers();
  }
  return $settings;
}
add_filter( 'wp_editor_settings', __NAMESPACE__.'\\extra_ticket_field_admin_helper_hook', 10, 2 );

/**
 * Display the extra fields to the admin
 */
function extra_ticket_field_admin_helpers() {
  $fields = array_keys( extra_fields() );
  $helpers = implode( "\n", array_map( function( $field_key ) {
    return sprintf( '<a href="#"><code>{%s}</code></a>', $field_key );
  }, $fields ) );
  ?>
  <p>
    Extra fields added through the theme:
    <span class="ticket-post-vars">
      <?= $helpers ?>
    </span>
  </p>
  <?php
}

function get_ticket_barcode_html( $ticket_id ) {

  // Get the image , it will automatically generate the text if required.
  // We don't need to store the result though.
  WC_Order_Barcodes()->get_barcode_image( $ticket_id, 'url' );

  $barcode_url = WC_Order_Barcodes()->barcode_url( $ticket_id );
  $barcode_text = get_post_meta( $ticket_id, '_barcode_text', true );

  $img = sprintf( '<img src="%s" />', $barcode_url );

  return "<div>
    {$img}<br>
    <code>{$barcode_text}</code>
    </div>";
}

/**
 * Add the new fields to the Box Office plugin
 */
function populate_printed_ticket_fields( $fields, $ticket_id, $product_id ) {

  $extra_field_keys = array_keys( extra_fields() );
  foreach ( $extra_field_keys as $field_key ) {
    $fields[$field_key] = [
      'label' => $field_key,
      'value' => extra_field_value( $ticket_id, $field_key ),
      'options' => [],
      'autofill' => 'none',
      'required' => 'no',
      'type' => null,
    ];
  }

  return $fields;
}
add_filter( 'woocommerce_box_office_get_ticket_fields', __NAMESPACE__.'\\populate_printed_ticket_fields', 10, 3 );


function bootstrap_field_vars( $vars ) {

  $vars['input_class'] .= ' form-control';
  $vars['before_field'] = '<div class="form-group">';
  $vars['after_field'] = '</div>';

  return $vars;
}
add_filter( 'wocommerce_box_office_input_field_template_vars', __NAMESPACE__.'\\bootstrap_field_vars', 10, 1 );
add_filter( 'wocommerce_box_office_option_field_template_vars', __NAMESPACE__.'\\bootstrap_field_vars', 10, 1 );


// Remove barcode from order email
if ( class_exists( 'WooCommerce_Order_Barcodes' ) ) {
  $wcob = \WooCommerce_Order_Barcodes::instance();
  remove_action( 'woocommerce_email_after_order_table', [ $wcob, 'get_email_barcode' ], 1 );
}


// Replace Box Office JS with our own patched version

function enqueue_scripts() {
  wp_register_script( 'woocommerce-box-office-frontend', \Roots\Sage\Assets\asset_path( 'scripts/box-office.js' ), [ 'jquery', 'imagesloaded' ], WCBO()->_version );
}
add_action( 'wp_enqueue_scripts', __NAMESPACE__.'\\enqueue_scripts', 1 );
